import traceback

from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from forest.services.resource_getter import ResourceGetter
from forest.services.resource_updater import ResourceUpdater
from forest.services.resource_remover import ResourceRemover

from forest.services.utils import jwt_verify
from forest.serializers.resource import ResourceSerializer


@csrf_exempt
@jwt_verify
def resource(request, model, r_id):
    try:
        json_api_data = {}
        if request.method == 'GET':
            getter = ResourceGetter(request, model, r_id)
            data = getter.perform()
            json_api_data = ResourceSerializer(model).serialize([data,], 1, single=True)
        elif request.method == 'PUT':
            updater = ResourceUpdater(request, model, r_id)
            data = updater.perform()
            json_api_data = ResourceSerializer(model).serialize([data,], 1, single=True)
        elif request.method == 'DELETE':
            remover = ResourceRemover(request, model, r_id)
            remover.perform()
            return HttpResponse(status=204)
    except Exception as err:
        traceback.print_exc()
        return HttpResponse(err, status=400)

    return JsonResponse(json_api_data, safe=False)




import jwt
import json
import datetime

from django.conf import settings
from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from forest.services.allowed_user_getter import get_allowed_users
from forest.services.utils import passwords_match

@csrf_exempt
def session(request):
    data = json.loads(request.body)
    allowedUsers = get_allowed_users(data.get('renderingId'))
    user = (u for u in allowedUsers if u['email'] == data['email']).next()
    if len(user) and passwords_match(data['password'], user['password']):
        payload = {
            'id': user['id'],
            'type': 'users',
            'data': {
                'email': user['email'],
                'first_name': user['first_name'],
                'last_name': user['last_name'],
                'teams': user['teams']
            },
            'exp': datetime.datetime.utcnow() + datetime.timedelta(days=14)
        }

        secret_key = settings.JWT_SECRET_KEY
        token = jwt.encode(payload, secret_key)
        return JsonResponse({ 'token': token })

    else:
        return HttpResponse(status=401)




from django.apps import apps


class ResourceGetter():
    def __init__(self, request, model, r_id):
        self.request = request
        self.params = request.GET
        self.model_name = model
        self.model = self._get_model_class()
        self.r_id = r_id

    def _get_model_class(self):
        models = apps.get_models()
        for model in models:
            if model.__name__.lower() == self.model_name.lower():
                return model

    def perform(self):
        return self.model.objects.get(pk=self.r_id)



import re
from operator import __or__ as OR

from django.db.models import ForeignKey
from django.db.models import Q

from forest.services import filter_generator
from forest.services import utils
from forest.generators import schemas


def _handle_limit(params):
    if params.get('page[size]'):
        return int(params['page[size]'])
    return 10

def _handle_offset(params, limit):
    if params.get('page[number]'):
        return (int(params['page[number]']) - 1) * limit
    return 0

def _handle_sort(params):
    if params.get('sort'):
        return params['sort'].replace('.', '__')

def _get_q_filter_key(model_name, association):
    model = utils.get_model_class(association)
    fields = model._meta.get_fields()
    for field in fields:
        if field.is_relation and field.related_model.__name__ == model_name:
            return field.name

def perform(request, model_name, r_id, association):
    model = utils.get_model_class(association)
    filter_key = _get_q_filter_key(model_name, association)
    q_filter = { '%s' % filter_key: r_id }
    #TODO: is this useful?
    related_fields = [f.name for f in model._meta.get_fields() if isinstance(f, ForeignKey)]
    params = request.GET
    limit = _handle_limit(params)
    offset = _handle_offset(params, limit)
    sort = _handle_sort(params)
    query = model.objects.select_related(*related_fields).filter(**q_filter)
    count = query.count()
    if sort:
        query = query.order_by(sort)

    return query[offset:offset+limit], count



import json

from django.core import serializers
from django.db.models import ForeignKey

from django.apps import apps
from forest.generators import schemas
from forest.serializers.serializer import Serializer as JSONApiSerializer

RESERVED_FIELDS = ['meta', 'attributes']


class ResourceSerializer():
    def __init__(self, model_name):
        self.model_name = model_name
        self.model = self.get_model_class(model_name)
        self.api_map = schemas.api_map.get()
        self.type_for_attribute = {}

    def _get_attributes(self):
        attributes = []
        for schema in self.api_map['data']:
            if schema.get('id') == self.model_name:
                attributes = schema.get('attributes')
        return attributes

    def _get_fields(self, attributes):
        return (x for x in attributes['fields'])

    def _fct_wrapper(self, field):
        """
        The relationship_links_fct function cannot be a closure, as the nested
        function looks up variables from the parent scope when executed,
        not when defined. It would therefore lose the "field" reference.
        """
        def relationship_links_fct(record, current, parent):
            return {
                'href': '/forest/%s/%s/%s' % (self.model_name, parent['id'],
                                              field)
            }
        return relationship_links_fct

    def _get_related_fields(self):
        return [f.name for f in self.model._meta.get_fields() if isinstance(f, ForeignKey)]

    def _get_model_fields(self, model_name):
        for model in self.api_map['data']:
            if model['id'] == model_name:
                return [f['field'] for f in model['attributes']['fields']]

    def _get_type_for_attribute(self):
        def inner_fct(string, attr):
            return self.type_for_attribute.get(string) or string
        return inner_fct

    def _get_options(self, queryset):
        attributes = self._get_attributes()
        fields = [x['field'] for x in attributes['fields']]
        for idx, field in enumerate(fields):
            if field in RESERVED_FIELDS:
                fields[idx] = field + ':namespace' + field
        options = {
            'attributes': fields,
            'key_for_attribute': (lambda x: x),
            'type_for_attribute': self._get_type_for_attribute()
        }
        related_fields = self._get_related_fields()
        for field in self._get_fields(attributes):
            included = field['field'].lower() in related_fields
            if field.get('reference'):
                options[field['field']] = { 'ref': 'id' }
                if not included:
                    options[field['field']]['included'] = False
                    options[field['field']]['ignore_relationship_data'] = True
                    options[field['field']]['relationship_links'] = {
                        'related': self._fct_wrapper(field['field'])
                    }
                else:
                    related_model = field['reference'].split('.')[0]
                    self.type_for_attribute[field['field']] = related_model
                    options[field['field']]['attributes'] = self._get_model_fields(related_model)
                    options[field['field']]['relationship_links'] = {
                        'related': self._fct_wrapper(related_model)
                    }


        return options

    def _format_records(self, queryset):
        attributes = self._get_attributes()
        attr_fields = [x['field'] for x in attributes['fields']]
        related_fields = self._get_related_fields()
        json_data = serializers.serialize('json', queryset)
        records = json.loads(json_data)

        data = []
        for index, rec in enumerate(records):
            rec['fields']['id'] = rec['pk']
            for field in attr_fields:
                if field not in rec['fields']:
                    rec['fields'][field] = []
            for field in related_fields:
                rel_obj = getattr(queryset[index], field)
                if not rel_obj:
                    continue
                ser_rel_obj = serializers.serialize('json', [rel_obj,])
                rel_data = json.loads(ser_rel_obj)[0]
                rel_data['fields']['id'] = rel_data['pk']
                rec['fields'][field] = rel_data['fields']
            data.append(rec['fields'])

        return data

    def get_model_class(self, model_name):
        models = apps.get_models()
        for model in models:
            if model.__name__.lower() == model_name.lower():
                return model

    def _set_meta(self, options, count):
        options['meta'] = { 'count': count }
        return options

    def serialize(self, queryset, count, single=False):
        name = self.model_name
        options = self._get_options(queryset)
        if not single:
            options = self._set_meta(options, count)
        resource_serializer = JSONApiSerializer(name, options)
        formatted_records = self._format_records(queryset)
        if single:
            formatted_records = formatted_records[0]
        return resource_serializer.serialize(formatted_records)




import inflection

def caserize(attribute, opts):
    attribute = inflection.underscore(attribute)
    option = opts.get('key_for_attribute')

    if option == 'dash-case':
        return attribute
    elif option == 'lisp-case':
        return attribute
    elif option == 'kebab-case':
        return inflection.dasherize(attribute)
    elif option == 'underscore_case':
        return attribute
    elif option == 'snake_case':
        return attribute
    elif option == 'CamelCase':
        return inflection.camelize(attribute)
    elif option == 'camelCase':
        return inflection.camelize(attribute, false)
    else:
        return inflection.dasherize(attribute)

def pluralize(type_):
    return inflection.pluralize(type_)



from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from forest.services.utils import jwt_verify
from forest.services import has_many_getter
from forest.serializers.resource import ResourceSerializer


@csrf_exempt
@jwt_verify
def association(request, model, r_id, association):
    data, count = has_many_getter.perform(request, model, r_id, association)
    json_api_data = ResourceSerializer(association).serialize(data, count)
    return JsonResponse(json_api_data, safe=False)



from django.db import models
from forest.services import filter_generator
from forest.services import utils

def _get_aggregate(params):
    return getattr(models, params.get('aggregate', 'Count'))

def _get_aggregate_field(params):
    return params.get('aggregate_field') or 'id'

def _get_filters(params):
    filters = []
    for item in params.get('filters', []):
        filters.append(filter_generator.get(item['field'], item['value']))

    all_filters = utils.merge_dicts(*filters)
    return all_filters

def perform(params, model_name):
    model = utils.get_model_class(model_name)
    aggregate_field = _get_aggregate_field(params)
    aggregate = _get_aggregate(params)
    filters = _get_filters(params)
    query = model.objects
    if filters:
        query = query.filter(**filters)
    data = query.aggregate(aggregate(aggregate_field))
    generated_field = "%s__%s" % (aggregate_field, params.get('aggregate').lower())
    count = data[generated_field]
    return {
        'value': {
            'countCurrent': count
        }
    }




from datetime import datetime, timedelta, date
from dateutil.relativedelta import relativedelta
from django.db import models
from forest.services import filter_generator
from forest.services import utils


def _get_filters(params):
    filters = []
    for item in params.get('filters', []):
        filters.append(filter_generator.get(item['field'], item['value']))

    all_filters = utils.merge_dicts(*filters)
    return all_filters


def _get_custom_select(params):
    select = {}
    group_by_date_field = params.get('group_by_date_field')
    time_range = params.get('time_range')
    if time_range == 'Day':
        select['day'] = 'date(%s)' % group_by_date_field
    elif time_range == 'Month':
        select['month'] = "extract(month FROM %s)" % group_by_date_field
        select['year'] = "extract(year FROM %s)" % group_by_date_field
    elif time_range == 'Year':
        select['year'] = "extract(year FROM %s)" % group_by_date_field
    elif time_range == 'Week':
        select['week'] = "date_trunc('week', %s)" % group_by_date_field

    return select


def _get_aggregate(params):
    aggregate = params.get('aggregate')
    if aggregate == 'Week':
        aggregate = 'WeekCount'
    return getattr(models, aggregate)


def _format_data(data, params):
    time_range = params.get('time_range')
    day = 1
    month = 1

    for item in data:
        if time_range == 'Day':
            date = _label_to_datetime(item.pop('day'))
            day = date.day
            month = date.month
            year = date.year
        elif time_range == 'Month':
            month = int(item.pop('month'))
            year = int(item.pop('year'))
        elif time_range == 'Year':
            year = int(item.pop('year'))
        elif time_range == 'Week':
            date = item.pop('week')
            day = date.day
            month = date.month
            year = date.year

        item['label'] = "%d-%02d-%02d" % (year, month, day)
        item['values'] = { 'value': item.pop('value') }

    _sort_data_by_date(data)
    _fill_empty_date(data, params)


def _sort_data_by_date(data):
    data.sort(key=lambda item:item['label'])


def _label_to_datetime(label):
    return datetime.strptime(label, '%Y-%m-%d')


def _datetime_to_label(date):
    return date.strftime('%Y-%m-%d')


def _find_by_label(data, label):
    return any(item['label'] == label for item in data)


def _fill_empty_date(data, params):
    time_range = params.get('time_range')
    relativedelta_param = { '%ss' % time_range.lower(): 1 }
    start_date = _label_to_datetime(data[0]['label'])
    end_date = _label_to_datetime(data[-1]['label'])
    while start_date < end_date:
        start_date = start_date + relativedelta(**relativedelta_param)
        label = _datetime_to_label(start_date)
        item = _find_by_label(data, label)
        if not item:
            data.append({ 'label': label, 'values': { 'value': 0 } })

    _sort_data_by_date(data)


def perform(params, model_name):
    model = utils.get_model_class(model_name)
    filters = _get_filters(params)
    select = _get_custom_select(params)
    values = select.keys()
    aggregate = _get_aggregate(params)
    query = model.objects.filter(**filters).extra(select=select).values(*values)
    query = query.annotate(value=aggregate(params.get('group_by_date_field')))
    data = list(query)
    _format_data(data, params)
    return { 'value': data }





import stripe
from django.conf import settings
from forest.services.utils import get_model_class


class StripePaymentsGetter():
    def __init__(self, params, model_name, r_id):
        self.model = get_model_class(model_name)
        self.r_id = r_id
        self.stripe = stripe
        self.stripe.api_key = settings.FOREST_STRIPE_API_KEY
        self.user_field = settings.FOREST_STRIPE_USER_FIELD
        self.params = params

    def _has_pagination(self):
        return self.params.get('page[number]')

    def _get_limit(self):
        if self._has_pagination():
            return int(self.params.get('page[size]', 10))
        else:
            return 10

    def _get_offset(self):
        if self._has_pagination():
            pages = int(self.params.get('page[number]'))
            return (pages - 1) * self._get_limit()
        else:
            return 0

    def perform(self):
        customer = self.model.objects.get(pk=self.r_id)
        query = {
            'limit': self.get_limit(),
            'offset': self.get_offset(),
            'source': { 'object': 'card' },
            'include[]': 'total_count'
        }

        if customer:
            query.customer = customer[self.user_field]

        charges = stripe.Charge.list(**query)
        for charge in charges:
            if customer:
                charge['data']['customer'] = customer
            else:
                query = { '%s' % self.user_field: r_id }
                customer = self.model.objects.filter(**query)
                charge['data']['customer'] = customer

        return [charges['total_count'], charges.data]



from django.db import models
from forest.services import filter_generator
from forest.services import utils

def _get_aggregate(params):
    return getattr(models, params.get('aggregate', 'Count'))

def _get_group_by(params):
    return params.get('group_by_field')

def _get_filters(params):
    filters = []
    for item in params.get('filters', []):
        filters.append(filter_generator.get(item['field'], item['value']))

    all_filters = utils.merge_dicts(*filters)
    return all_filters

def perform(params, model_name):
    model = utils.get_model_class(model_name)
    group_by_field = _get_group_by(params)
    aggregate = _get_aggregate(params)
    filters = _get_filters(params)
    query = model.objects
    query = query.filter(**filters)
    query = query.values(group_by_field)
    query = query.annotate(value=aggregate(group_by_field))

    # The following cast to list is necessary to override the "remaining
    # elements truncated" limitation of django queryset result. Maybe there's a
    # better way.
    # TODO: Should we set a limit? Would make sense for pie charts.

    data = list(query.order_by('-value'))
    for item in data:
        item['key'] = item.pop(group_by_field)

    return { 'value': data }




from django.db.models import Q
from django.db.models import Lookup
from django.db.models.fields import Field

from forest.services import operator_date_interval_parser as date_parser

class NotEqual(Lookup):
    lookup_name = 'ne'

    def as_sql(self, qn, connection):
        lhs, lhs_params = self.process_lhs(qn, connection)
        rhs, rhs_params = self.process_rhs(qn, connection)
        params = lhs_params + rhs_params
        return '%s <> %s' % (lhs, rhs), params

Field.register_lookup(NotEqual)

def get(field_name, value):

    if not len(value):
        return

    if value == 'true':
        return {field_name: True}

    elif value == 'false':
        return {field_name: False}

    elif value[0] == '!':
        value = value[1:]
        return {"%s__ne" % field_name: value}

    elif value[0] == '<':
        value = value[1:]

        if date_parser.is_interval_date_value(value):
            start, end = date_parser.get_interval_date_filter(value)
            return {"%s__range" % field_name: [start, end]}

        return {"%s__lt" % field_name: value}

    elif value[0] == '>':
        value = value[1:]

        if date_parser.is_interval_date_value(value):
            start, end = date_parser.get_interval_date_filter(value)
            return {"%s__range" % field_name: [start, end]}

        return {"%s__gt" % field_name: value}

    elif value[0] == '*' and value[-1] == '*':
        value = value[1:-1]
        return {"%s__icontains" % field_name: value}

    elif value[0] == '*':
        value = value[1:]
        return {"%s__endswith" % field_name: value}

    elif value[-1] == '*':
        value = value[:-1]
        return {"%s__starswith" % field_name: value}

    elif value == '$present':
        return {"%s__isnull" % field_name: False}

    elif value == '$blank':
        return {"%s__isnull" % field_name: True}

    else:
        return {field_name: value}





import traceback

from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from forest.services.utils import jwt_verify
from forest.services.resources_getter import ResourcesGetter
from forest.services.resource_creator import ResourceCreator
from forest.serializers.resource import ResourceSerializer


@csrf_exempt
@jwt_verify
def resources(request, model):
    try:
        json_api_data = {}
        if request.method == 'GET':
            getter = ResourcesGetter(request, model)
            data = getter.perform()
            count = getter.count
            json_api_data = ResourceSerializer(model).serialize(data, count)
        elif request.method == 'POST':
            creator = ResourceCreator(request, model)
            data = creator.perform()
            json_api_data = ResourceSerializer(model).serialize([data,], 1, single=True)
    except Exception as err:
        traceback.print_exc()
        return HttpResponse(err, status=400)

    return JsonResponse(json_api_data, safe=False)




import json
import dateutil.parser
from forest.services.utils import merge_dicts, get_model_class
from forest.generators import schemas


class ResourceCreator():
    def __init__(self, request, model_name):
        self.body = json.loads(request.body)
        self.model_name = model_name
        self.model = get_model_class(model_name)

    def perform(self):
        kw_params_list = []
        for field, value in self.body['data'].get('attributes', {}).iteritems():
            if value:
                field_type = schemas.api_map.get_type(self.model_name, field)
                if field_type == 'Date':
                    value = dateutil.parser.parse(value)
                elif field_type == 'Number':
                    value = float(value)
                kw_params_list.append({'%s' % field: value})


        for k, v in self.body['data'].get('relationships', {}).iteritems():
            if not isinstance(v['data'], list):
                kw_params_list.append({"%s_id" % k: v['data'].get('id')})

        kw_params = merge_dicts(*kw_params_list)
        item = self.model(**kw_params)
        item.save()
        return item




import os
import requests
from django.conf import settings

def get_allowed_users(rendering_id):
    url = os.getenv('FOREST_URL', getattr(settings, 'FOREST_URL',
         'https://forestadmin-server.herokuapp.com'))
    url += '/forest/renderings/%s/allowed-users' % rendering_id
    secret_key = os.getenv('FOREST_SECRET_KEY', settings.FOREST_SECRET_KEY)
    headers = {
        'forest-secret-key':  secret_key,
        'Content-Type': 'application/json'
    }
    r = requests.get(url, headers=headers)
    allowedUsers = []
    for data in r.json()['data']:
        user = data['attributes']
        user['id'] = data['id']
        allowedUsers.append(user)
    return allowedUsers



import jwt
import bcrypt
from django.http import HttpResponse, JsonResponse
from django.apps import apps
from django.conf import settings


def merge_dicts(*dict_args):
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

def get_model_class(model_name):
    models = apps.get_models()
    for model in models:
        if model.__name__.lower() == model_name.lower():
            return model

def passwords_match(plain, hashed):
    return bcrypt.hashpw(str(plain), str(hashed)) == str(hashed)

def jwt_get_user_id_from_payload_handler(payload):
    return payload.get('id')

def jwt_verify(view_func):
    def _wrapped_view_func(request, *args, **kwargs):
        if hasattr(request, 'META') and request.META.get('HTTP_AUTHORIZATION'):
            splitted_token = request.META['HTTP_AUTHORIZATION'].split()
            if len(splitted_token):
                token = splitted_token[1]
        try:
            payload = jwt.decode(token, settings.JWT_SECRET_KEY, True)
            return view_func(request, *args, **kwargs)
        except jwt.DecodeError as err:
            return HttpResponse(status=401)

    return _wrapped_view_func



import json
import traceback
from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt

from forest.services.utils import jwt_verify
from forest.services import value_stat_getter
from forest.services import pie_stat_getter
from forest.services import line_stat_getter
from forest.serializers.stat import StatSerializer


@csrf_exempt
def stat(request, model):
    try:
        if request.method == 'POST':
            payload = json.loads(request.body)
            if payload.get('type') == 'Value':
                data = value_stat_getter.perform(payload, model)
                json_api_data = StatSerializer().serialize(data)
                return JsonResponse(json_api_data, safe=False)
            elif payload.get('type') == 'Pie':
                data = pie_stat_getter.perform(payload, model)
                json_api_data = StatSerializer().serialize(data)
                return JsonResponse(json_api_data, safe=False)
            elif payload.get('type') == 'Line':
                data = line_stat_getter.perform(payload, model)
                json_api_data = StatSerializer().serialize(data)
                return JsonResponse(json_api_data, safe=False)
    except Exception as err:
        traceback.print_exc()
        return HttpResponse(err, status=400)

    return HttpResponse(status=204)



import re

from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

PERIODS = [
    'yesterday',
    'lastWeek',
    'last2Weeks',
    'lastMonth',
    'last3Months',
    'lastYear'
]

PERIODS_VALUES = {
    'days': 'day',
    'weeks': 'isoWeek',
    'months': 'month',
    'years': 'year'
}

PERIOD_LAST_X_DAYS_REGEX = re.compile(r'^last(\d+)days$')

def is_interval_date_value(value):
    if value in PERIODS:
        return True

    match = PERIOD_LAST_X_DAYS_REGEX.match(value)
    if match:
        return True

    return False

def reset_time(date):
    return date.replace(hour=0, minute=0, second=0, microsecond=0)

def get_interval_date_filter(value):
    if not is_interval_date_value(value):
        return

    now = datetime.utcnow()

    match = PERIOD_LAST_X_DAYS_REGEX.match(value)
    if match:
        start = now - timedelta(days=match.group(1))
        return (start,)

    start = end = None

    if value == 'yesterday':
        start = now - timedelta(days=1)
        start = reset_time(start)
        end = now
        end = reset_time(end)

    elif value == 'lastWeek':
        start = now - timedelta(weeks=1)
        start = start - timedelta(days=start.weekday())
        start = reset_time(start)
        end = now - timedelta(days=6-now.weekday())
        end = reset_time(end)

    elif value == 'last2Weeks':
        start = now - timedelta(weeks=2)
        start = start - timedelta(days=start.weekday())
        start = reset_time(start)
        end = now - timedelta(days=6-now.weekday())
        end = reset_time(end)

    elif value == 'lastMonth':
        start = now - relativedelta(months=1)
        start = start.replace(start.year, start.month, 1)
        start = reset_time(start)
        end = now.replace(now.year, now.month, 1)
        end = reset_time(end)

    elif value == 'last3Month':
        start = now - relativedelta(months=3)
        start = start.replace(start.year, start.month, 1)
        start = reset_time(start)
        end = now.replace(now.year, now.month, 1)
        end = reset_time(end)

    elif value == 'lastYear':
        start = now - relativedelta(years=1)
        start = start.replace(start.year, 1, 1)
        start = reset_time(start)
        end = now.replace(now.year, 1, 1)
        end = reset_time(end)

    return (start, end)




import json
import dateutil.parser
from forest.services.utils import merge_dicts, get_model_class
from forest.generators import schemas


class ResourceRemover():
    def __init__(self, request, model_name, r_id):
        self.model = get_model_class(model_name)
        self.r_id = r_id

    def perform(self):
        item = self.model.objects.get(pk=self.r_id)
        item.delete()




from django.conf.urls import url
from forest import views
from forest.views import empty, session, resources, resource, association, stat

urlpatterns = [
    url(r'^$', empty),
    url(r'^sessions$', session),
    url(r'(?i)^(?P<model>[a-z]*)$', resources),
    url(r'(?i)^(?P<model>[a-z]*)/(?P<r_id>[0-9]+)$', resource),
    url(r'(?i)^(?P<model>[a-z]*)/(?P<r_id>[0-9]+)/(?P<association>[a-z]*)$', association),
    url(r'(?i)^stats/(?P<model>[a-z]*)$', stat),
]



import importlib
import pkgutil
from django.conf import settings
from django.db import models
from django.apps import apps
from forest.serializers.serializer import Serializer as JSONApiSerializer

FIELDS_TYPE_MAPPING = {
    'AutoField': 'Number',
    'BigIntegerField': 'Number',
    'BinaryField': 'Number',
    'BooleanField': 'Boolean',
    'CharField': 'String',
    'CommaSeparatedIntegerField': 'String',
    'DateField': 'Date',
    'DateTimeField': 'Date',
    'DecimalField': 'Number',
    # TODO: not sure about DurationField
    # https://docs.djangoproject.com/en/1.9/ref/models/fields/#durationfield
    'DurationField': 'Number',
    'EmailField': 'String',
    # TODO: not sure about FileField
    # https://docs.djangoproject.com/en/1.9/ref/models/fields/#filefield
    'FileField': 'String',
    'FilePathField': 'String',
    'FloatField': 'Number',
    # TODO: not sure about ImageField
    # https://docs.djangoproject.com/en/1.9/ref/models/fields/#imagefield
    'ImageField': 'String',
    'IntegerField': 'Number',
    'GenericIPAddressField': 'String',
    # TODO: not sure about NullBooleanField
    # https://docs.djangoproject.com/en/1.9/ref/models/fields/#nullbooleanfield
    'NullBooleanField': 'Boolean',
    'PositiveIntegerField': 'Number',
    'PositiveSmallIntegerField': 'Number',
    'SlugField': 'String',
    'SmallIntegerField': 'Number',
    'TextField': 'String',
    'TimeField': 'Date',
    'URLField': 'String',
    'UUIDField': 'String',
}


class ApiMap():
    def __init__(self):
        self.apimap = {}
        self.actions = []
        self.init_actions()
        self.smart_collections = []
        self.init_smart_collections()

    def get(self):
        return self.apimap

    def get_models(self, app_names):
        models = []
        for app_name in app_names:
            splitted_app_name = app_name.split('.')
            if len(splitted_app_name):
                if splitted_app_name[-1] == '*':
                    for model in settings.INSTALLED_APPS:
                        if model.startswith('.'.join(splitted_app_name[:-1])):
                            models.append(model)
                    return self.get_models(models)
                else:
                    for model in apps.get_app_config(splitted_app_name[-1]).get_models():
                        models.append(model)
        return models

    def generate(self):
        data = []
        models = self.get_models(settings.FOREST_APPS)
        for model in models:
            data.append(self.get_schema(model))

        data += self.smart_collections
        self.apimap = self.serialize(data)
        return self.apimap

    def get_custom_module_path(self):
        if len(settings.FOREST_APPS):
            return settings.FOREST_APPS[0].split('.')[0]

    def init_actions(self):
        """Imports actions from FOREST_APP_NAME/forest/actions folder. This
        does not import recursively.
        """
        try:
            mod_path = "%s.forest.actions" % self.get_custom_module_path()
            actions_module = importlib.import_module(mod_path)
            for loader, name, ispkg in pkgutil.iter_modules(actions_module.__path__):
                action_module = loader.find_module(name).load_module(name)
                if hasattr(action_module, 'ACTION'):
                    self.actions.append(action_module.ACTION)
        except ImportError as err:
            pass

    def init_smart_collections(self):
        """Imports actions from FOREST_APP_NAME/forest/smart_collections
        folder. This does not import recursively.
        """
        try:
            mod_path = "%s.forest.smart_collections" % self.get_custom_module_path()
            sc_module = importlib.import_module(mod_path)
            for loader, name, ispkg in pkgutil.iter_modules(sc_module.__path__):
                sc_module = loader.find_module(name).load_module(name)
                if hasattr(sc_module, 'SMART_COLLECTION'):
                    self.smart_collections.append(sc_module.SMART_COLLECTION)
        except ImportError as err:
            pass

    def get_actions(self, model_name):
        actions = []
        for action in self.actions:
            if action.get('collection') == model_name:
                actions.append(action)
        return actions

    def get_schema(self, model):
        schema = {
            'name': model.__name__,
            'fields': self.get_model_fields(model),
            'only-for-relationship': None,
            'is-virtual': None,
            'is-read-only': False,
            'is-searchable': True,
            'actions': self.get_actions(model.__name__)
        }
        return schema

    def serialize(self, apimap):
        options = {
            'id': 'name',
            'attributes': ['name', 'displayName', 'paginationType', 'icon',
                           'fields', 'actions', 'onlyForRelationships',
                           'isVirtual', 'isReadOnly'],
            'fields': {
                'attributes': ['field', 'displayName', 'type', 'enums',
                               'collection_name', 'reference', 'column',
                               'isSearchable', 'widget', 'integration']
            },
            'actions': {
                'ref': 'name',
                'attributes': ['name', 'endpoint', 'httpMethod', 'fields']
            },
            'meta': {
                'liana': 'forest-django',
                'liana_version': '0.0.1'
            }
        }
        ser = JSONApiSerializer('collections', options)
        ser_apimap = ser.serialize(apimap)
        return ser_apimap

    def get_model_fields(self, model):
        fields = []
        for field in model._meta.get_fields():
            internal_type = field.get_internal_type()
            schema = {
                'field': field.name,
                'type': FIELDS_TYPE_MAPPING.get(internal_type, 'String')
            }
            if not field.is_relation:
                fields.append(schema)
            else:
                ref = field.related_model.__name__
                pk = field.related_model._meta.pk.name

                if field.one_to_many or field.many_to_many:
                    schema['field'] = ref
                    schema['type'] = ['Number']

                elif field.many_to_one:
                    ref = field.related_model.__name__

                schema['reference'] = "%s.%s" % (ref, pk)
                fields.append(schema)

        return fields

    def get_model_schema(self, model_name):
        schema = filter(lambda x: x['id'] == model_name, self.apimap['data'])
        if len(schema):
            return schema[0]

    def get_field_from_model(self, model_name, field_model):
        schema = self.get_model_schema(model_name)
        if schema is None:
            return
        for field in schema['attributes']['fields']:
            if field.get('reference', '').split('.')[0] == field_model:
                return field['field']

    def get_type(self, model_name, field_name):
        schema = self.get_model_schema(model_name)
        if schema is None:
            return

        for field in schema['attributes']['fields']:
            if field['field'] == field_name:
                return field['type']

api_map = ApiMap()



from pydash import _
from serializer_utils import SerializerUtils


class Serializer(object):
    def __init__(self, collection_name, opts):
        self.payload = {}
        self.collection_name = collection_name
        self.opts = opts

    def serialize(self, records):
        if self.opts.get('top_level_links'):
            self.payload['links'] = \
            self._get_links(self.opts['top_level_links'], records)

        if self.opts.get('meta'):
            self.payload['meta'] = self.opts['meta']

        if _.is_list(records):
            return self._collection(records)
        else:
            return self._resource(records)

    def _get_links(self, links, records):
        def value_mapper(value):
            if _.is_function(value):
                return value(records)
            else:
                return value
        return _.map_values(links, value_mapper)

    def _collection(self, records):
        self.payload['data'] = []

        for rec in records:
            su = SerializerUtils(self.collection_name, rec, self.payload,
                                 self.opts)
            self.payload['data'].append(su.perform())

        return self.payload

    def _resource(self, records):
        self.payload['data'] = SerializerUtils(self.collection_name,
                                               records, self.payload,
                                               self.opts).perform()
        return self.payload



import inflector
from pydash import _


class SerializerUtils(object):
    def __init__(self, collection_name, record, payload, opts):
        self.collection_name = collection_name
        self.record = record
        self.payload = payload
        self.opts = opts

    def key_for_attribute(self, attribute):
        def transform_callback(acc, value, key):
            if self._is_complex_type(value):
                acc[self.key_for_attribute(key)] = self.key_for_attribute(value)
            else:
                acc[self.key_for_attribute(key)] = value
            return acc, value, key

        def map_function(item):
            if self._is_complex_type(item):
                return self.key_for_attribute(item)
            else:
                return item

        if _.is_dict(attribute):
            return _.transform(attribute, transform_callback)


        elif _.is_list(attribute):
            map(map_function, attribute)

        else:
            if _.is_function(self.opts.get('key_for_attribute')):
                return self.opts['key_for_attribute'](attribute)
            else:
                caserized = inflector.caserize(attribute, self.opts)
                return caserized

    def get_id(self):
        return self.opts.get('id', 'id')

    def _is_complex_type(self, obj):
        return _.is_list(obj) or _.is_dict(obj)

    def get_ref(self, current, item, opts):
        if _.is_function(opts.get('ref')):
            return opts['ref'](current, item)
        elif opts.get('ref') is True:
            if _.is_list(item):
                return map(lambda val: str(val), item)
            elif item:
                return str(item)
        elif isinstance(item, dict) and item.get(opts.get('ref')):
            return str(item[opts['ref']])

    def get_type(self, string, attr_val):
        type_ = None
        attr_val = attr_val or {}
        if _.is_function(self.opts.get('type_for_attribute')):
            type_ = self.opts['type_for_attribute'](string, attr_val)

        if (self.opts.get('pluralize_type') is None or self.opts.get('pluralize_type')) and type_ is None:
            type_ = inflector.pluralize(string)

        if type_ is None:
            type_ = string

        return type_

    def get_links(self, current, links, dest):
        def map_function(item):
            if _.is_function(item):
                return item(self.record, current, dest)
            else:
                return item
        return _.map_values(links, map_function)

    def get_meta(self, current, meta):
        def map_function(item):
            if _.is_function(item):
                return item(self.record, current)
            else:
                return item

        return _.map_values(meta, map_function)

    def pick(self, obj, attributes):
        def map_function(value, key):
            return self.key_for_attribute(key)

        return _.map_keys(_.pick(obj, attributes), map_function)

    def is_compound_document_included(self, included, item):
        return _.find(self.payload.get('included', {}), {
            'id': item.get('id'),
            'type': item.get('type')
        })

    def push_to_included(self, dest, include):
        if not self.is_compound_document_included(dest, include):
            if not dest.get('included'):
                dest['included'] = []
            dest['included'].append(include)

    def serialize(self, dest, current, attribute, opts):
        data = None

        if isinstance(opts, dict) and opts.get('ref'):
            if not dest.get('relationships'):
                dest['relationships'] = {}

            def map_current(item):
                return self.serialize_ref(item, current, attribute, opts)

            if _.is_list(current.get(attribute)):
                data = map(map_current, current[attribute])

            else:
                data = self.serialize_ref(current[attribute], current,
                                         attribute, opts)

            dest['relationships'][self.key_for_attribute(attribute)] = {}
            if not opts.get('ignore_relationship_data'):
                dest['relationships'][self.key_for_attribute(attribute)]['data'] = data

            if opts.get('relationship_links'):
                dest['relationships'][self.key_for_attribute(attribute)]['links'] = \
                    self.get_links(current[attribute], opts['relationship_links'], dest)

            if opts.get('relationship_meta'):
                dest['relationships'][self.key_for_attribute(attribute)]['meta'] = \
                    self.get_meta(current['attribute'], opts['relationship_meta'])
        else:
            if _.is_list(current[attribute]):
                if len(current[attribute]) and _.is_dict(current[attribute][0]):
                    def map_current(item):
                        return self.serialize_nested(item, current, attribute,
                                                    opts)
                    data = map(map_current, current[attribute])
                else:
                    data = current[attribute]

                dest['attributes'][self.key_for_attribute(attribute)] = data
            elif _.is_dict(current[attribute]):
                data = self.serialize_nested(current[attribute], current,
                                            attribute, opts)
                dest['attributes'][self.key_for_attribute(attribute)] = data
            else:
                dest['attributes'][self.key_for_attribute(attribute)] = current[attribute]

    def serialize_ref(self, dest, current, attribute, opts):
        id_ = self.get_ref(current, dest, opts)
        type_ = self.get_type(attribute, dest)

        relationships = []
        included_attrs = []

        if opts.get('attributes'):
            relationships = filter(lambda x: opts.get(x), opts['attributes'])
            included_attrs = filter(lambda x: opts.get(x) is None, opts['attributes'])

        included = { 'type': type_, 'id': id_ }
        if included_attrs:
            included['attributes'] = self.pick(dest, included_attrs)

        for rel in relationships:
            if self._is_complex_type(dest[rel]):
                self.serialize(included, dest, rel, opts[rel])

        if included_attrs and (opts.get('included') is None or
                               opts.get('included') is True):
            if opts.get('included_links'):
                included['links'] = self.get_links(dest, opts['included_links'])

            if id_ is not None:
                self.push_to_included(self.payload, included)

        if id_ is not None:
            return { 'type': type_, 'id': id_ }
        else:
            return None

    def serialize_nested(self, dest, current, attribute, opts):

        embeds = []
        attributes = []

        if opts and opts.get('attributes'):
            embeds = filter(lambda x: opts.get(x), opts['attributes'])
            attributes = filter(lambda x: opts.get(x) is None, opts['attributes'])
        else:
            attributes = _.keys(dest)

        ret = {}
        if attributes:
            ret['attributes'] = self.pick(dest, attributes)

        for embed in embeds:
            if self._is_complex_type(dest[embed]):
                self.serialize(ret, dest, embed, opts[embed])

        return ret['attributes']

    def perform(self):

        if self.record is None:
            return None

        data = {
            'type': self.get_type(self.collection_name, self.record),
            'id': str(self.record[self.get_id()])
        }

        if self.opts.get('data_links'):
            data['links'] = self.get_links(self.record, self.opts['data_links'], None)

        for attr in self.opts.get('attributes', []):
            splitted_attributes = attr.split(':')
            if len(splitted_attributes) and splitted_attributes[0] in self.record:
                if not data.get('attributes'):
                    data['attributes'] = {}

                attribute_map = attr;
                if len(splitted_attributes) > 1:
                    attr = splitted_attributes[0]
                    attribute_map = splitted_attributes[1]

                self.serialize(data, self.record, attr, self.opts.get(attribute_map))

        return data



from django.http import HttpResponse, JsonResponse
from django.views.decorators.csrf import csrf_exempt
from forest.services.stripe_payments_getter import StripePaymentsGetter
from forest.serializers.stripe import StripePaymentsSerializer

@csrf_exempt
def stripe_payments(request):
    getter = StripePaymentsGetter(request)
    data, count = getter.perform()
    json_api_data = StripePaymentsSerializer().serialize(data, count)
    return JsonResponse(json_api_data, safe=False)


@csrf_exempt
def stripe_refund(request):
    return HttpResponse(status=204)

@csrf_exempt
def stripe_invoices(request):
    return HttpResponse(status=204)

@csrf_exempt
def stripe_cards(request):
    return HttpResponse(status=204)





import json
import requests
import os

from django.conf import settings
from django.apps import AppConfig
from forest.generators.schemas import api_map
from forest.services.utils import get_model_class

class ForestConfig(AppConfig):
    name = 'forest'
    verbose_name = "My Forest Connector"
    def ready(self):
        apimap = api_map.generate()
        url = os.getenv('FOREST_URL', getattr(settings, 'FOREST_URL',
             'https://forestadmin-server.herokuapp.com'))
        url += '/forest/apimaps'
        secret_key = os.getenv('FOREST_SECRET_KEY', settings.FOREST_SECRET_KEY)
        headers = {
            'forest-secret-key':  secret_key,
            'Content-Type': 'application/json'
        }
        req = requests.post(url, data=json.dumps(apimap), headers=headers)



import uuid
from forest.serializers.serializer import Serializer as JSONApiSerializer

class StatSerializer():
    def serialize(self, stat):
        stat['id'] = uuid.uuid4()
        options = {
            'attributes': ['value'],
            'key_for_attribute': (lambda x: x)
        }
        ser = JSONApiSerializer('stat', options)
        return ser.serialize(stat)



import re
from operator import __or__ as OR

from django.db.models import ForeignKey
from django.db.models import Q

from forest.services import filter_generator
from forest.services import utils


class ResourcesGetter():
    def __init__(self, request, model, r_id=None, association=None):
        self.request = request
        self.params = request.GET
        self.model = utils.get_model_class(model)
        self.count = 0
        self.r_id = r_id
        self.association = association

    def _get_limit(self):
        if self.params.get('page[size]'):
            return int(self.params['page[size]'])
        return 10

    def _get_offset(self):
        if self.params.get('page[number]'):
            return (int(self.params['page[number]']) - 1) * self._get_limit()
        return 0

    def _get_sort(self):
        if self.params.get('sort'):
            return self.params['sort'].replace('.', '__')

    def _handle_search_param(self, params):
        pass

    def is_filter_field_valid(self, field):
        try:
            return field in [f.name for f in self.model._meta.get_fields()]
        except Exception as err:
            import pdb; pdb.set_trace()

    def _merge_dicts(self, *dict_args):
        result = {}
        for dictionary in dict_args:
            result.update(dictionary)
        return result

    def _handle_filter_param(self):
        filters = []

        for k,v in self.params.iteritems():
            if k.startswith('filter['):
                match = re.match(r'filter\[(.*)\]', k)
                if match:
                    filter_field = match.group(1)
                    if self.is_filter_field_valid(filter_field):
                        filters.append(filter_generator.get(filter_field, v))

        all_filters = self._merge_dicts(*filters)
        return all_filters

    def _handle_search_param(self):
        search_param = self.params.get('search')
        if not search_param:
            return

        fields = [f.name for f in self.model._meta.get_fields() if not f.is_relation]
        q_list = [Q(**{'%s__icontains' % f : search_param}) for f in fields]
        related = [f for f in self.model._meta.get_fields() if isinstance(f, ForeignKey)]
        for rel in related:
            model = rel.related_model
            r_fields = [f.name for f in model._meta.get_fields() if not f.is_relation]
            q_list += [Q(**{'%s__%s__icontains' % (rel.name, r_field): search_param}) for r_field in r_fields]
        search_filter = reduce(OR, q_list)

        return search_filter

    def _get_related_fields(self):
        return [f.name for f in self.model._meta.get_fields() if isinstance(f, ForeignKey)]

    def _build_query(self):
        related_fields = self._get_related_fields()
        sort = self._get_sort()
        filters = self._handle_filter_param()
        search_filter = self._handle_search_param()

        query = self.model.objects
        if filters:
            query = query.filter(**filters)
        elif search_filter:
            query = query.filter(search_filter)
        if sort:
            query = query.order_by(sort)
        if related_fields:
            query = query.select_related(*related_fields)

        return query


    def perform(self):
        self.count = self._build_query().count()
        offset = self._get_offset()
        limit = self._get_limit()
        data = self._build_query()[offset:offset+limit]
        return data
